/*
* -cr_beg-
* ********************************************************************
* IBM Confidential
* OCO Source Materials
* 5724-U38, 5724-U39, 5724-U48, 5655-V17, 5655-V19, 5655-V35, 5655-V37
* (C) Copyright IBM Corp. 2012, 2013
* The source code for this program is not published or otherwise
* divested of its trade secrets, irrespective of what has
* been deposited with the U.S. Copyright Office.
* ********************************************************************
* -cr_end-
*/

/*!\file
 *\verbatim
 *--------------------------------------------------------------------*
 * File:    OptimDistReportJNI.c
 * Desc:     Interface between Optim Distributed C/C++ reporting and Java
 *  JAXB classes
 *
 * Function: This file contains the following functions:
 *
 *
 *
 * Author:   RAP
 * Created:  6/16/13
 *
 *
 *--------------------------------------------------------------------*
\endverbatim */

#include <stdio.h>
#include <string.h>
#include <jni.h>
#include <ctype.h>
#include <stdarg.h>
#include "OptimReportingContext.h"

static const char *classPathOption = "-Djava.class.path=";
static const char *classBasePath = "com/ibm/nex/xml/schema";
static const char *reportingJarName = "optimdistreport.jar";
static const char *OptimReportingContextClassName = "com/ibm/nex/xml/schema/OptimReportingContext";
static const char *OptimReportHelperClassName = "com/ibm/nex/xml/schema/reporting/OptimReportHelper";
static const char *OptimReportingMessageClassName = "com/ibm/nex/xml/schema/OptimReportingMessage";
static const char *OptimReportingNameValuePairClassName = "com/ibm/nex/xml/schema/OptimReportingNameValuePair";
static const char *OptimReportingNumberPairClassName = "com/ibm/nex/xml/schema/OptimReportingNumberPair";
static const char *OptimReportingEntityClassName = "com/ibm/nex/xml/schema/OptimReportingEntity";
static const char *OptimReportingRelationshipClassName = "com/ibm/nex/xml/schema/OptimReportingRelationship";
static const char *OptimReportingStatisticsClassName = "com/ibm/nex/xml/schema/OptimReportingStatistics";

static const int RC_SUCCESS = 0;
static const int RC_WARNING = 4;
static const int RC_ERROR   = 8;
static const int RC_FATAL   = 12;

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#if defined(WIN32)
static const char pathSepChar = '\\';
static const char *pathSepString = "\\";
#else
static const char pathSepChar = '/';
static const char *pathSepString = "/";
#endif

static char *getBasename(char *);
static char *truncateBasename(char *);
static void error(char *, ...);
static void debug(char *, ...);

static jstring convertWcharToJString(wchar_t *, JNIEnv *env);

static unsigned char debugFlag = 1;

#if !defined(Convert32To16)
static int Convert32To16(const unsigned int *Utf32,
                     int Utf32Chars,
                     unsigned short *Utf16,
                     int Utf16Size,
                     int *pOutUtf16Size);
#endif

#if !defined(strrchr)
char * strrchr(const char *inStr, int c)
{
	char *retPtr = (char *)NULL;
	char *cPtr = inStr;
	int cVal;

	if (inStr == (char *)NULL) {
		return ((char *)NULL);
	}

	while ((cVal = *cPtr++) != '\0')
	{
		if (cVal == c) {
			retPtr = cPtr - 1; // back the pointer down, since we incremented
		}
	}
	return (retPtr);
}
#endif

/*!\verbatim
 *--------------------------------------------------------------------*
 * Func:     initJvm
 *
 * Purpose:  Initializes JVM and activates reporting jar.
 *
 * Parms:
 *     JavaVM** pointer to java VM pointer -- java VM pointer populated
 *              by call
 *	   JNIEnv** pointer to JNI environment pointer -- JNI environment
 *	            pointer populated by call
 *
 * Returns:
 *		Return code from Java method invoked.
 *
 *--------------------------------------------------------------------*
\endverbatim */

int initJvm(JavaVM **javaVM, JNIEnv **env)
{
    JavaVMOption options[1];
    char optionString[1024];
    char programName[1024]; // Overkill, but consistent with optionString
    char *baseProgramDir;
    JavaVMInitArgs initArgs;
    int rc;
    jclass stringClass;

    (void)strcpy(optionString, classPathOption);
//    (void) strcat(optionString, argv[0]);
//    (void)strcpy(programName,argv[0]);
//    baseProgramDir = truncateBasename(programName);
    baseProgramDir = ".";
    (void)strcat(optionString,baseProgramDir);
    (void)strcat(optionString,pathSepString);
    (void)strcat(optionString,reportingJarName);
    options[0].optionString = optionString;

    initArgs.version = JNI_VERSION_1_6;
    initArgs.nOptions = 1;
    initArgs.options = options;
    initArgs.ignoreUnrecognized = JNI_TRUE;
    debug("JNI options are:  %s\n", options[0].optionString);
    rc = JNI_CreateJavaVM(javaVM, (void**)env, &initArgs);
    if (rc < 0) {
        error("Error: Unable to create JVM.\n");
        return RC_FATAL;
    }
    debug("Finding String class in init\n");
    stringClass = (**env)->FindClass(*env,"java/lang/String");
    if (stringClass == NULL) {
          error("Error: Unable to find String\n");
    } else {
    	debug("Found String!!\n");
    }

    return rc;
}

int createReport(jobject javaContext, JavaVM *javaVM, JNIEnv *env)
{
	int rc = RC_SUCCESS;
	jclass jOptimReportHelperClass;
	jmethodID jReportCreate;
	jclass stringClass;

    debug("Finding String class in createReport\n");
    stringClass = (*env)->FindClass(env,"java/lang/String");
    if (stringClass == NULL) {
          error("Error: Unable to find String\n");
    } else {
    	debug("Found String!!\n");
    }

	/* Call OptimReportHelper.OptimReportHelper.createProcessAndDisplayReportFromContext with the populated context */
	jOptimReportHelperClass = (*env)->FindClass(env, OptimReportHelperClassName);
    if((*env)->ExceptionCheck(env))
    {
        (*env)->ExceptionDescribe(env);
        (*env)->ExceptionClear(env);
    }
    if (jOptimReportHelperClass == NULL) {
        error("Error: Unable to find report helper class. %s\n", OptimReportHelperClassName);
        return RC_FATAL;
    }
    jReportCreate = (*env)->GetStaticMethodID(env, jOptimReportHelperClass,"createProcessAndDisplayReportFromContext", "(Lcom/ibm/nex/xml/schema/OptimReportingContext;)I");
    if((*env)->ExceptionCheck(env))
    {
        (*env)->ExceptionDescribe(env);
        (*env)->ExceptionClear(env);
    }
    if (jReportCreate == NULL) {
        error("Error: Unable to find OptimReportHelper.createProcessAndDisplayReportFromContext() method.");
        return RC_FATAL;
    }

    rc = (*env)->CallStaticIntMethod(env, jOptimReportHelperClass, jReportCreate, javaContext);
    if((*env)->ExceptionCheck(env))
    {
        (*env)->ExceptionDescribe(env);
        (*env)->ExceptionClear(env);
    }
	return rc;
}

#define POPULATE_STRING_FIELD(fieldName, setterName, javaClass, javaObject, container) \
	    if (container fieldName != (wchar_t *)NULL) { \
	    	javaStringValue = convertWcharToJString(container fieldName, env); \
	    	if (javaStringValue != (jstring) NULL) { \
	        	javaSetter = (*env)->GetMethodID(env, javaClass, #setterName, "(Ljava/lang/String;)V"); \
	        	if (javaSetter != (jmethodID) NULL) { \
	        		(*env)->CallVoidMethod(env, javaObject, javaSetter, javaStringValue); \
	                /*(*env)->DeleteLocalRef(env, javaSetter);*/ \
	        	} else { \
	        		error("Could not find method %s\n", #setterName); \
	        		(*env)->DeleteGlobalRef(env, javaStringValue); \
	        		return RC_FATAL; \
	        	} \
	        	(*env)->DeleteGlobalRef(env, javaStringValue); \
	    	} else { \
	    		error("Could not obtain string value for field %s\n", #fieldName); \
	    		return RC_FATAL; \
	    	} \
	    }

#define POPULATE_NUMBER_FIELD(fieldName, setterName, typeSig, castType, javaClass, javaObject, container) \
	    { \
		    javaSetter = (*env)->GetMethodID(env, javaClass, #setterName, typeSig); \
		    if (javaSetter != (jmethodID) NULL) { \
				(*env)->CallVoidMethod(env, javaObject, javaSetter, (castType)container fieldName); \
		        /*(*env)->DeleteLocalRef(env, javaSetter);*/ \
		    } else { \
				error("Could not find method %s\n", #fieldName); \
				return RC_FATAL; \
		    } \
		}

#define POPULATE_INT_FIELD(fieldName, setterName, javaClass, javaObject, container) POPULATE_NUMBER_FIELD(fieldName, setterName, "(I)V", jint, javaClass, javaObject, container)
#define POPULATE_LONG_FIELD(fieldName, setterName, javaClass, javaObject, container) POPULATE_NUMBER_FIELD(fieldName, setterName, "(J)V", jlong, javaClass, javaObject, container)
#define POPULATE_BOOLEAN_FIELD(fieldName, setterName, javaClass, javaObject, container) POPULATE_NUMBER_FIELD(fieldName, setterName, "(Z)V", jboolean, javaClass, javaObject, container)
#define POPULATE_DOUBLE_FIELD(fieldName, setterName, javaClass, javaObject, container) POPULATE_NUMBER_FIELD(fieldName, setterName, "(D)V", jdouble, javaClass, javaObject, container)

#define POPULATE_CONTEXT_INT_FIELD(fieldName, setterName) POPULATE_NUMBER_FIELD(fieldName, setterName, "(I)V", jint, jContextCls, localContext, context->)
#define POPULATE_CONTEXT_LONG_FIELD(fieldName, setterName) POPULATE_NUMBER_FIELD(fieldName, setterName, "(J)V", jlong, jContextCls, localContext, context->)
#define POPULATE_CONTEXT_BOOLEAN_FIELD(fieldName, setterName) POPULATE_NUMBER_FIELD(fieldName, setterName, "(Z)V", jboolean, jContextCls, localContext, context->)
#define POPULATE_CONTEXT_DOUBLE_FIELD(fieldName, setterName) POPULATE_NUMBER_FIELD(fieldName, setterName, "(D)V", jdouble, jContextCls, localContext, context->)
#define POPULATE_CONTEXT_STRING_FIELD(fieldName, setterName) POPULATE_STRING_FIELD(fieldName, setterName, jContextCls, localContext, context->)


int populateJavaContext(OptimReportingContext_t *context, jobject *javaContext, JavaVM *javaVM, JNIEnv *env)
{
	int rc = RC_SUCCESS;
	jclass jContextCls;
	jmethodID jContextConstructor;
	jobject localContext;
	jmethodID javaSetter;
	jstring  javaStringValue;
	jstring	 javaStringValue2;
	jclass stringClass;
	jclass jArrayListClass;
    jmethodID jArrayListAddMID;
    jmethodID jArrayListGetMID;
    jmethodID jListGetter;
    jobject   jList;
    jboolean  addStatus;
    jclass	  jListElementClass;
    jmethodID jListElementConstructor;
    jobject   jListElement;
	int idx;

    debug("Finding String class in populate\n");
    stringClass = (*env)->FindClass(env,"java/lang/String");
    if (stringClass == NULL) {
          error("Error: Unable to find String\n");
    } else {
    	debug("Found String!!\n");
    }

	jContextCls = (*env)->FindClass(env, OptimReportingContextClassName);
    if (jContextCls == NULL) {
        error("Error: Unable to find report context class. %s\n", OptimReportingContextClassName);
        return RC_FATAL;
    }
    jContextConstructor = (*env)->GetMethodID(env, jContextCls, "<init>", "()V");
    if (jContextConstructor == NULL) {
        error("Error: Unable to find OptimReportingContext constructor.");
        return RC_FATAL;
    }

    localContext = (*env)->NewObject(env, jContextCls, jContextConstructor);
    if (localContext == NULL) {
        error("Error: Unable to instantiate loader.\n");
        return RC_FATAL;
    }
    *javaContext = (*env)->NewGlobalRef(env,localContext);

    if (context->reportFile != (wchar_t *)NULL) {
    	javaStringValue = convertWcharToJString(context->reportFile, env);
    	if (javaStringValue != (jstring) NULL) {
        	javaSetter = (*env)->GetMethodID(env, jContextCls, "setReportFileName", "(Ljava/lang/String;)V");
        	if (javaSetter != (jmethodID) NULL) {
        		(*env)->CallVoidMethod(env, localContext, javaSetter, javaStringValue);
                /*(*env)->DeleteLocalRef(env, javaSetter);*/
        	} else {
        		error("Could not find method %s\n", "setReportFileName");
        		(*env)->DeleteGlobalRef(env, javaStringValue);
        		return RC_FATAL;
        	}
        	(*env)->DeleteGlobalRef(env, javaStringValue);
    	} else {
    		error("Could not obtain string value for field %s\n", "reportFile");
    		return RC_FATAL;
    	}
    }
   /* POPULATE_CONTEXT_STRING_FIELD(reportFile, setReportFileName) */             /* reportFile */
    POPULATE_CONTEXT_STRING_FIELD(htmlReportFile, setHtmlReportFileName)     /* htmlReportFile */

    /* It would be nice to do some fancy reflection thing here.  BUT, for now, brute force will do */
    /* AccessDefinitionModificationDate */
    POPULATE_CONTEXT_LONG_FIELD(errorCount, setErrorCount)
    POPULATE_CONTEXT_LONG_FIELD(infoCount, setInfoCount)
    POPULATE_CONTEXT_LONG_FIELD(warningCount, setWarningCount)
    POPULATE_CONTEXT_LONG_FIELD(timeElapsed, setTimeElapsed)
    POPULATE_CONTEXT_LONG_FIELD(totalRows, setTotalRows)
    POPULATE_CONTEXT_LONG_FIELD(totalRowsFirstPass, setTotalRowsFirstPass)
    POPULATE_CONTEXT_LONG_FIELD(totalRowsInError, setTotalRowsInError)
    POPULATE_CONTEXT_LONG_FIELD(totalTables, setTotalTables)

    POPULATE_CONTEXT_STRING_FIELD(optimDirectoryName, setOptimDirectoryName)
    POPULATE_CONTEXT_STRING_FIELD(optimDirectoryDBMSType, setOptimDirectoryDBMSType)
    POPULATE_CONTEXT_STRING_FIELD(optimDirectoryDBMSVendor, setOptimDirectoryDBMSVendor)
    POPULATE_CONTEXT_STRING_FIELD(optimDirectoryDBMSVersion, setOptimDirectoryDBMSVersion)
    POPULATE_CONTEXT_STRING_FIELD(optimSolutionName, setOptimSolutionName)
    POPULATE_CONTEXT_STRING_FIELD(optimSolutionVersion, setOptimSolutionVersion)
    POPULATE_CONTEXT_STRING_FIELD(optimRuntimeName, setOptimRuntimeName)
    POPULATE_CONTEXT_STRING_FIELD(optimRuntimeVersion, setOptimRuntimeVersion)
    POPULATE_CONTEXT_STRING_FIELD(clientUserId, setClientUserId)
    POPULATE_CONTEXT_STRING_FIELD(requestName, setRequestName)
    POPULATE_CONTEXT_STRING_FIELD(serverName, setServerName)
    POPULATE_CONTEXT_STRING_FIELD(serverUserId, setServerUserId)
    POPULATE_CONTEXT_STRING_FIELD(timeStarted, setTimeStarted)
    POPULATE_CONTEXT_STRING_FIELD(timeFinished, setTimeFinished)
    POPULATE_CONTEXT_STRING_FIELD(accessDefinitionName, setAccessDefinitionName)
    POPULATE_CONTEXT_STRING_FIELD(accessDefinitionModificationDate, setAccessDefinitionModificationDate)
    POPULATE_CONTEXT_STRING_FIELD(extractFile, setExtractFile)
    POPULATE_CONTEXT_STRING_FIELD(compareFile, setCompareFile)
    POPULATE_CONTEXT_STRING_FIELD(controlFile, setControlFile)

    POPULATE_CONTEXT_INT_FIELD(accessDefinitionType, setAccessDefinitionType)
    POPULATE_CONTEXT_INT_FIELD(reportType, setReportType)

    /* Now populate the lists... First -- configure JNI for handling ArrayList */
    debug("Going after Messages list\n");
    jArrayListClass = (*env)->FindClass(env, "java/util/ArrayList");
    jArrayListAddMID = (*env)->GetMethodID(env, jArrayListClass, "add", "(Ljava/lang/Object;)Z");
    jArrayListGetMID = (*env)->GetMethodID(env, jArrayListClass, "get", "(I)Ljava/lang/Object;");
    /* Get the messages list */
    jListGetter = (*env)->GetMethodID(env, jContextCls, "getMessages", "()Ljava/util/List;");
    jList = (jobject) NULL; /* Clear out jList */
	if (jListGetter != (jmethodID) NULL) {
		jList = (*env)->CallObjectMethod(env, localContext, jListGetter);
        /*(*env)->DeleteLocalRef(env, javaSetter);*/
	} else {
		error("Could not find method %s\n", "getMessages");
		return RC_FATAL;
	}
	/* Set things up to create OptimReportingMessage objects */
	debug("LIST ops set up, now set up Message object stuff\n");
	jListElementClass = (*env)->FindClass(env, OptimReportingMessageClassName);
    if (jListElementClass == NULL) {
        error("Error: Unable to find list element class. %s\n", OptimReportingMessageClassName);
        return RC_FATAL;
    }
    jListElementConstructor = (*env)->GetMethodID(env, jListElementClass, "<init>", "(JLjava/lang/String;I)V");
    if (jListElementConstructor == NULL) {
        error("Error: Unable to find %s constructor.", OptimReportingMessageClassName);
        return RC_FATAL;
    }

	debug("Got message object stuff done. Populating\n");
    for (idx = 0; idx < context->numMessages; idx ++) {
    	/* Do String construction/conversion */
    	javaStringValue = convertWcharToJString(context->messages[idx].message, env);
    	if (javaStringValue == (jstring) NULL) {
    		error("Could not obtain string value for field %s\n", "reportFile");
    		return RC_FATAL;
    	}
    	/* Create list element */
        jListElement = (*env)->NewObject(env, jListElementClass, jListElementConstructor,(jlong)context->messages[idx].id, javaStringValue , (jint)context->messages[idx].type);
        if (jListElement == NULL) {
            error("Error: Unable to instantiate element for %s.\n", OptimReportingMessageClassName);
            return RC_FATAL;
        }
        /* Add element to list and clean up references */
        addStatus = (*env)->CallBooleanMethod(env,jList,jArrayListAddMID, jListElement);
        if((*env)->ExceptionCheck(env)) {
            (*env)->ExceptionDescribe(env);
            (*env)->ExceptionClear(env);
        	error("Error: add() of element to list %s generated Exception.\n", OptimReportingMessageClassName);
        	return RC_FATAL;
        }
        if (!addStatus) {
        	error("Error: add() of element to list %s failed.\n", OptimReportingMessageClassName);
        	return RC_FATAL;
        }
        (*env)->DeleteLocalRef(env, jListElement);
       	(*env)->DeleteGlobalRef(env, javaStringValue);
    }
   	(*env)->DeleteGlobalRef(env, jList);
   	(*env)->DeleteGlobalRef(env, jListElementClass);
    jList = (jobject) NULL; /* Clear out jList */
    jListElementClass = (jclass) NULL;

    debug ("Done with message list \n");
    /* Do the Request Options Next */
    debug ("Beginning Request Options\n");
    /* Get the request options list */
    jListGetter = (*env)->GetMethodID(env, jContextCls, "getRequestOptions", "()Ljava/util/List;");
	if (jListGetter != (jmethodID) NULL) {
		jList = (*env)->CallObjectMethod(env, localContext, jListGetter);
        /*(*env)->DeleteLocalRef(env, javaSetter);*/
	} else {
		error("Could not find method %s\n", "getMessages");
		return RC_FATAL;
	}
	/* Set things up to create OptimReportingNameValuePair objects */
	debug("LIST ops set up, now set up NameValuePair object stuff\n");
	jListElementClass = (*env)->FindClass(env, OptimReportingNameValuePairClassName);
    if (jListElementClass == NULL) {
        error("Error: Unable to find list element class. %s\n", OptimReportingNameValuePairClassName);
        return RC_FATAL;
    }
    jListElementConstructor = (*env)->GetMethodID(env, jListElementClass, "<init>", "(Ljava/lang/String;Ljava/lang/String;)V");
    if (jListElementConstructor == NULL) {
        error("Error: Unable to find %s constructor.", OptimReportingNameValuePairClassName);
        return RC_FATAL;
    }

	debug("Got request option object stuff done. Populating\n");
    for (idx = 0; idx < context->numRequestOptions; idx ++) {
    	/* Do String construction/conversion */
    	javaStringValue = convertWcharToJString(context->requestOptions[idx].name, env);
    	if (javaStringValue == (jstring) NULL) {
    		error("Could not obtain string value for field %s\n", "name");
    		return RC_FATAL;
    	}
    	javaStringValue2 = convertWcharToJString(context->requestOptions[idx].value, env);
    	if (javaStringValue2 == (jstring) NULL) {
    		error("Could not obtain string value for field %s\n", "name");
    		return RC_FATAL;
    	}
    	/* Create list element */
        jListElement = (*env)->NewObject(env, jListElementClass, jListElementConstructor, javaStringValue, javaStringValue2);
        if (jListElement == NULL) {
            error("Error: Unable to instantiate element for %s.\n", OptimReportingNameValuePairClassName);
            return RC_FATAL;
        }
        /* Add element to list and clean up references */
        addStatus = (*env)->CallBooleanMethod(env,jList,jArrayListAddMID, jListElement);
        if((*env)->ExceptionCheck(env)) {
            (*env)->ExceptionDescribe(env);
            (*env)->ExceptionClear(env);
        	error("Error: add() of element to list %s generated Exception.\n", OptimReportingNameValuePairClassName);
        	return RC_FATAL;
        }
        if (!addStatus) {
        	error("Error: add() of element to list %s failed.\n", OptimReportingNameValuePairClassName);
        	return RC_FATAL;
        }
        (*env)->DeleteLocalRef(env, jListElement);
       	(*env)->DeleteGlobalRef(env, javaStringValue);
       	(*env)->DeleteGlobalRef(env, javaStringValue2);
    }
   	(*env)->DeleteGlobalRef(env, jList);
   	(*env)->DeleteGlobalRef(env, jListElementClass);
    jList = (jobject) NULL; /* Clear out jList */
    jListElementClass = (jclass) NULL;
    debug("Done with RequestOptions\n");

    /* Now do Extended Objects */
    debug ("Beginning ExtendedObjects\n");
    /* Get the request options list */
    jListGetter = (*env)->GetMethodID(env, jContextCls, "getExtendedObjects", "()Ljava/util/List;");
	if (jListGetter != (jmethodID) NULL) {
		jList = (*env)->CallObjectMethod(env, localContext, jListGetter);
        /*(*env)->DeleteLocalRef(env, javaSetter);*/
	} else {
		error("Could not find method %s\n", "getExtendedObjects");
		return RC_FATAL;
	}
	/* Set things up to create OptimReportingNameValuePair objects */
	debug("LIST ops set up, now set up NumberPair object stuff\n");
	jListElementClass = (*env)->FindClass(env, OptimReportingNumberPairClassName);
    if (jListElementClass == NULL) {
        error("Error: Unable to find list element class. %s\n", OptimReportingNumberPairClassName);
        return RC_FATAL;
    }
    jListElementConstructor = (*env)->GetMethodID(env, jListElementClass, "<init>", "(JJ)V");
    if (jListElementConstructor == NULL) {
        error("Error: Unable to find %s constructor.", OptimReportingNumberPairClassName);
        return RC_FATAL;
    }

	debug("Got request option object stuff done. Populating\n");
    for (idx = 0; idx < context->numExtendedObjects; idx ++) {
     	/* Create list element */
        jListElement = (*env)->NewObject(env, jListElementClass, jListElementConstructor, (jlong)context->extendedObjects[idx].number1, (jlong)context->extendedObjects[idx].number2);
        if (jListElement == NULL) {
            error("Error: Unable to instantiate element for %s.\n", OptimReportingNumberPairClassName);
            return RC_FATAL;
        }
        /* Add element to list and clean up references */
        addStatus = (*env)->CallBooleanMethod(env,jList,jArrayListAddMID, jListElement);
        if((*env)->ExceptionCheck(env)) {
            (*env)->ExceptionDescribe(env);
            (*env)->ExceptionClear(env);
        	error("Error: add() of element to list %s generated Exception.\n", OptimReportingNumberPairClassName);
        	return RC_FATAL;
        }
        if (!addStatus) {
        	error("Error: add() of element to list %s failed.\n", OptimReportingNameValuePairClassName);
        	return RC_FATAL;
        }
        (*env)->DeleteLocalRef(env, jListElement);
    }
   	(*env)->DeleteGlobalRef(env, jList);
   	(*env)->DeleteGlobalRef(env, jListElementClass);
    jList = (jobject) NULL; /* Clear out jList */
    jListElementClass = (jclass) NULL;
    debug("Done with ExtendedObjects\n");

    /* CREATE ENTITIES LIST */
    debug ("Beginning Entities\n");
    jListGetter = (*env)->GetMethodID(env, jContextCls, "getEntities", "()Ljava/util/List;");
	if (jListGetter != (jmethodID) NULL) {
		jList = (*env)->CallObjectMethod(env, localContext, jListGetter);
        /*(*env)->DeleteLocalRef(env, javaSetter);*/
	} else {
		error("Could not find method %s\n", "getEntities");
		return RC_FATAL;
	}
	/* Set things up to create OptimReportingEntity objects */
	debug("LIST ops set up, now set up OptimReportingEntity object stuff\n");
	jListElementClass = (*env)->FindClass(env, OptimReportingEntityClassName);
    if (jListElementClass == NULL) {
        error("Error: Unable to find list element class. %s\n", OptimReportingEntityClassName);
        return RC_FATAL;
    }
    jListElementConstructor = (*env)->GetMethodID(env, jListElementClass, "<init>", "()V");
    if (jListElementConstructor == NULL) {
        error("Error: Unable to find %s constructor.", OptimReportingEntityClassName);
        return RC_FATAL;
    }

	debug("Got entity object stuff done. Populating\n");
    for (idx = 0; idx < context->numEntities; idx ++) {
     	/* Create list element */
        jListElement = (*env)->NewObject(env, jListElementClass, jListElementConstructor);
        if (jListElement == NULL) {
            error("Error: Unable to instantiate element for %s.\n", OptimReportingEntityClassName);
            return RC_FATAL;
        }
        /* Add element to list and clean up references */
        addStatus = (*env)->CallBooleanMethod(env,jList,jArrayListAddMID, jListElement);
        if((*env)->ExceptionCheck(env)) {
            (*env)->ExceptionDescribe(env);
            (*env)->ExceptionClear(env);
        	error("Error: add() of element to list %s generated Exception.\n", OptimReportingEntityClassName);
        	return RC_FATAL;
        }
        if (!addStatus) {
        	error("Error: add() of element to list %s failed.\n", OptimReportingEntityClassName);
        	return RC_FATAL;
        }
        /* NOW, POPULATE FIELDS IN LIST ELEMENT */
        POPULATE_LONG_FIELD(compareDifferent, setCompareDifferent, jListElementClass, jListElement, context->entities[idx]. )
        POPULATE_LONG_FIELD(compareDuplicate, setCompareDuplicate, jListElementClass, jListElement, context->entities[idx]. )
        POPULATE_LONG_FIELD(compareEqual, setCompareEqual, jListElementClass, jListElement, context->entities[idx]. )
        POPULATE_LONG_FIELD(compareUnmatched, setCompareUnmatched, jListElementClass, jListElement, context->entities[idx]. )
        POPULATE_STRING_FIELD(convertCsvFileName, setConvertCsvFileName, jListElementClass, jListElement, context->entities[idx]. )
        POPULATE_LONG_FIELD(convertCsvLimitFileCount, setConvertCsvLimitFileCount, jListElementClass, jListElement, context->entities[idx]. )
        POPULATE_LONG_FIELD(convertCsvLimitFileSize, setConvertCsvLimitFileSize, jListElementClass, jListElement, context->entities[idx]. )
        POPULATE_LONG_FIELD(convertCsvLimitRowCount, setConvertCsvLimitRowCount, jListElementClass, jListElement, context->entities[idx]. )
        POPULATE_BOOLEAN_FIELD(deleteNotFound, setDeleteNotFound, jListElementClass, jListElement, context->entities[idx]. )
        POPULATE_DOUBLE_FIELD(extractAverageRowLength, setExtractAverageRowLength, jListElementClass, jListElement, context->entities[idx]. )
        POPULATE_BOOLEAN_FIELD(extractCompressionEnabled, setExtractCompressionEnabled, jListElementClass, jListElement, context->entities[idx]. )
        POPULATE_INT_FIELD(extractCompressionValue, setExtractCompressionValue, jListElementClass, jListElement, context->entities[idx]. )
        POPULATE_BOOLEAN_FIELD(extractLob, setExtractLob, jListElementClass, jListElement, context->entities[idx]. )
        POPULATE_DOUBLE_FIELD(extractMaxLobLength, setExtractMaxLobLength, jListElementClass, jListElement, context->entities[idx]. )
        POPULATE_LONG_FIELD(failed, setFailed, jListElementClass, jListElement, context->entities[idx]. )
        POPULATE_LONG_FIELD(insertInserted, setInsertInserted, jListElementClass, jListElement, context->entities[idx]. )
        POPULATE_LONG_FIELD(insertUpdated, setInsertUpdated, jListElementClass, jListElement, context->entities[idx]. )
        POPULATE_STRING_FIELD(name, setName, jListElementClass, jListElement, context->entities[idx]. )
        POPULATE_LONG_FIELD(recordCount, setRecordCount, jListElementClass, jListElement, context->entities[idx]. )
        POPULATE_LONG_FIELD(skipped, setSkipped, jListElementClass, jListElement, context->entities[idx]. )
        POPULATE_STRING_FIELD(status, setStatus, jListElementClass, jListElement, context->entities[idx]. )
        POPULATE_INT_FIELD(type, setType, jListElementClass, jListElement, context->entities[idx]. )
        (*env)->DeleteLocalRef(env, jListElement);
    }
   	(*env)->DeleteGlobalRef(env, jList);
   	(*env)->DeleteGlobalRef(env, jListElementClass);
    jList = (jobject) NULL; /* Clear out jList */
    jListElementClass = (jclass) NULL;
    debug("Done with Entities\n");

    /* Relationships */
    debug ("Beginning Relationships\n");
    jListGetter = (*env)->GetMethodID(env, jContextCls, "getRelationships", "()Ljava/util/List;");
	if (jListGetter != (jmethodID) NULL) {
		jList = (*env)->CallObjectMethod(env, localContext, jListGetter);
        /*(*env)->DeleteLocalRef(env, javaSetter);*/
	} else {
		error("Could not find method %s\n", "getRelationships");
		return RC_FATAL;
	}
	/* Set things up to create OptimReportingRelationship objects */
	debug("LIST ops set up, now set up OptimReportingRelationship object stuff\n");
	jListElementClass = (*env)->FindClass(env, OptimReportingRelationshipClassName);
    if (jListElementClass == NULL) {
        error("Error: Unable to find list element class. %s\n", OptimReportingRelationshipClassName);
        return RC_FATAL;
    }
    jListElementConstructor = (*env)->GetMethodID(env, jListElementClass, "<init>", "()V");
    if (jListElementConstructor == NULL) {
        error("Error: Unable to find %s constructor.", OptimReportingRelationshipClassName);
        return RC_FATAL;
    }

	debug("Got relationship object stuff done. Populating\n");
    for (idx = 0; idx < context->numRelationships; idx ++) {
     	/* Create list element */
        jListElement = (*env)->NewObject(env, jListElementClass, jListElementConstructor);
        if (jListElement == NULL) {
            error("Error: Unable to instantiate element for %s.\n", OptimReportingRelationshipClassName);
            return RC_FATAL;
        }
        /* Add element to list and clean up references */
        addStatus = (*env)->CallBooleanMethod(env,jList,jArrayListAddMID, jListElement);
        if((*env)->ExceptionCheck(env)) {
            (*env)->ExceptionDescribe(env);
            (*env)->ExceptionClear(env);
        	error("Error: add() of element to list %s generated Exception.\n", OptimReportingRelationshipClassName);
        	return RC_FATAL;
        }
        if (!addStatus) {
        	error("Error: add() of element to list %s failed.\n", OptimReportingRelationshipClassName);
        	return RC_FATAL;
        }
        /* NOW, POPULATE FIELDS IN LIST ELEMENT */
        POPULATE_INT_FIELD(childAccessType, setChildAccessType, jListElementClass, jListElement, context->relationships[idx]. )
        POPULATE_INT_FIELD(childKeyLimit, setChildKeyLimit, jListElementClass, jListElement, context->relationships[idx]. )
        POPULATE_STRING_FIELD(childName, setChildName, jListElementClass, jListElement, context->relationships[idx]. )
        POPULATE_INT_FIELD(parentAccessType, setParentAccessType, jListElementClass, jListElement, context->relationships[idx]. )
        POPULATE_INT_FIELD(parentKeyLimit, setParentKeyLimit, jListElementClass, jListElement, context->relationships[idx]. )
        POPULATE_STRING_FIELD(parentName, setParentName, jListElementClass, jListElement, context->relationships[idx]. )
        POPULATE_STRING_FIELD(relationshipName, setRelationshipName, jListElementClass, jListElement, context->relationships[idx]. )
        (*env)->DeleteLocalRef(env, jListElement);
    }
   	(*env)->DeleteGlobalRef(env, jList);
   	(*env)->DeleteGlobalRef(env, jListElementClass);
    jList = (jobject) NULL; /* Clear out jList */
    jListElementClass = (jclass) NULL;
    debug("Done with Relationships\n");

    /* NOW, do STATISTICS */
    debug ("Beginning Statistics\n");
    jListGetter = (*env)->GetMethodID(env, jContextCls, "getStatistics", "()Ljava/util/List;");
	if (jListGetter != (jmethodID) NULL) {
		jList = (*env)->CallObjectMethod(env, localContext, jListGetter);
        /*(*env)->DeleteLocalRef(env, javaSetter);*/
	} else {
		error("Could not find method %s\n", "getStatistics");
		return RC_FATAL;
	}
	/* Set things up to create OptimReportingStatistics objects */
	debug("LIST ops set up, now set up OptimReportingStatistics object stuff\n");
	jListElementClass = (*env)->FindClass(env, OptimReportingStatisticsClassName);
    if (jListElementClass == NULL) {
        error("Error: Unable to find list element class. %s\n", OptimReportingStatisticsClassName);
        return RC_FATAL;
    }
    jListElementConstructor = (*env)->GetMethodID(env, jListElementClass, "<init>", "()V");
    if (jListElementConstructor == NULL) {
        error("Error: Unable to find %s constructor.", OptimReportingStatisticsClassName);
        return RC_FATAL;
    }

	debug("Got statistics object stuff done. Populating\n");
    for (idx = 0; idx < context->numStatistics; idx ++) {
     	/* Create list element */
        jListElement = (*env)->NewObject(env, jListElementClass, jListElementConstructor);
        if (jListElement == NULL) {
            error("Error: Unable to instantiate element for %s.\n", OptimReportingStatisticsClassName);
            return RC_FATAL;
        }
        /* Add element to list and clean up references */
        addStatus = (*env)->CallBooleanMethod(env,jList,jArrayListAddMID, jListElement);
        if((*env)->ExceptionCheck(env)) {
            (*env)->ExceptionDescribe(env);
            (*env)->ExceptionClear(env);
        	error("Error: add() of element to list %s generated Exception.\n", OptimReportingStatisticsClassName);
        	return RC_FATAL;
        }
        if (!addStatus) {
        	error("Error: add() of element to list %s failed.\n", OptimReportingStatisticsClassName);
        	return RC_FATAL;
        }
        /* NOW, POPULATE FIELDS IN LIST ELEMENT */
        POPULATE_INT_FIELD(childStrategy, setChildStrategy, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_INT_FIELD(columnCount, setColumnCount, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_BOOLEAN_FIELD(cycle, setCycle, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_INT_FIELD(databaseConnections, setDatabaseConnections, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_INT_FIELD(dbAccessType, setDbAccessType, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_LONG_FIELD(dbDataPerSecond, setDbDataPerSecond, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_INT_FIELD(dbKeysPer, setDbKeysPer, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_STRING_FIELD(dbmsVendor, setDbmsVendor, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_STRING_FIELD(dbmsVersion, setDbmsVersion, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_INT_FIELD(dbOpenCursor, setDbOpenCursor, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_LONG_FIELD(dbProcessTime, setDbProcessTime, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_LONG_FIELD(dbRowsFetched, setDbRowsFetched, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_LONG_FIELD(dbRowsPerSecond, setDbRowsPerSecond, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_LONG_FIELD(dbRowsWritten, setDbRowsWritten, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_LONG_FIELD(dbTimeInAct, setDbTimeInAct, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_LONG_FIELD(dbTimeInDB, setDbTimeInDB, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_LONG_FIELD(estimatedRowCount, setEstimatedRowCount, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_LONG_FIELD(fetchBuffer, setFetchBuffer, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_LONG_FIELD(foreignKeyIndex, setForeignKeyIndex, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_LONG_FIELD(foreignKeyNoIndex, setForeignKeyNoIndex, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_BOOLEAN_FIELD(lobs, setLobs, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_INT_FIELD(method, setMethod, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_STRING_FIELD(name, setName, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_INT_FIELD(parentStrategy, setParentStrategy, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_LONG_FIELD(primaryKeyIndex, setPrimaryKeyIndex, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_LONG_FIELD(primaryKeyNoIndex, setPrimaryKeyNoIndex, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_INT_FIELD(relAccess, setRelAccess, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_INT_FIELD(relDb2Cost, setRelDb2Cost, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_INT_FIELD(relDb2ScanCost, setRelDb2ScanCost, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_INT_FIELD(relDirection, setRelDirection, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_BOOLEAN_FIELD(relIndex, setRelIndex, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_INT_FIELD(relKeyLength, setRelKeyLength, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_INT_FIELD(relKeyPerCur, setRelKeyPerCur, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_INT_FIELD(relLookupKey, setRelLookupKey, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_STRING_FIELD(relName, setRelName, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_STRING_FIELD(relSql, setRelSql, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_INT_FIELD(relType, setRelType, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_INT_FIELD(rowLength, setRowLength, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_BOOLEAN_FIELD(selectUniqueRecord, setSelectUniqueRecord, jListElementClass, jListElement, context->statistics[idx]. )
        POPULATE_INT_FIELD(stepNumber, setStepNumber, jListElementClass, jListElement, context->statistics[idx]. )
        (*env)->DeleteLocalRef(env, jListElement);
    }
   	(*env)->DeleteGlobalRef(env, jList);
   	(*env)->DeleteGlobalRef(env, jListElementClass);
    jList = (jobject) NULL; /* Clear out jList */
    jListElementClass = (jclass) NULL;
    debug("Done with Statistics\n");



    (*env)->DeleteLocalRef(env, jContextCls);
    (*env)->DeleteLocalRef(env, stringClass);
    (*env)->DeleteLocalRef(env, jArrayListClass);
	return rc;
}

int outputReport(OptimReportingContext_t *context) {
    JavaVM *javaVM;
    JNIEnv *env;
    int rc;
    jobject javaContext;

    /* Check context */
    if (context == (OptimReportingContext_t *) NULL) {
    	error("Context cannot be NULL.");
    	return RC_FATAL;
    }
    debug("Initializing JVM.\n");

    rc = initJvm(&javaVM, &env);
    if (rc > RC_WARNING) {
    	error("Problem creating JVM.  Consult logs.\n");
    	return rc;
    }
    /* JVM Created. Need to terminate JVM whenever we bail*/
    /* Now, we need to create an OptimReportingContext in Java and populate it */
    debug("populating Java context \n");
    rc = populateJavaContext(context, &javaContext, javaVM, env);
    if (rc > RC_WARNING) {
    	error("Problem populating Java OptimReportingContext. Consult logs.\n");
        (*javaVM)->DestroyJavaVM(javaVM);
        return rc;
    }
    debug("Context populated, outputting report\n");
    /* With the context populated, we pass it to OptimReportHelper.createProcessAndDisplayReportFromContext() */
    rc = createReport(javaContext, javaVM, env);
    if (rc > RC_WARNING) {
    	error("Problem creating report. Consult logs.\n");
    	/* Fall through to destroy JVM */
    }

    debug("Reporting done RC= %d\n", rc);
    /* Terminate JVM */
    (*javaVM)->DestroyJavaVM(javaVM);
    return rc;
}

#ifndef WCSLEN
#define WCSLEN(wstr) wcslen((wstr))
#endif
#ifndef WCSNCPY
#define WCSNCPY(wto,wfrom,wlen) wcsncpy((wto),(wfrom),(wlen))
#endif

/* The caller must free the string */
jstring convertWcharToJString(wchar_t *wSource, JNIEnv *env) {
	jstring retStr;
	jstring jDest;
	jchar tmpBuf[256];
	size_t wSourceNumChars;
	jchar *javaTmpStg = (jchar *)NULL;
	jint jStgUsed;

	if (wSource == (wchar_t *) NULL) {
		return ((jstring) NULL);
	}

	wSourceNumChars = WCSLEN(wSource);
	if ((wSourceNumChars+1)*sizeof(jchar) <= sizeof(tmpBuf)) {
		javaTmpStg = tmpBuf;
	} else {
		javaTmpStg = (jchar *)malloc((wSourceNumChars + 1) * sizeof(jchar));
	}
	memset(javaTmpStg, 0, (wSourceNumChars + 1) * sizeof(jchar));

	/* WIDE 4-byte wchar_t */
    if (sizeof(wchar_t) == 4) {
    	Convert32To16((unsigned int *)wSource, wSourceNumChars, javaTmpStg, wSourceNumChars, &jStgUsed);
    } else { /* 2-byte wchar_t -- we accommodate no other choices */
        WCSNCPY((wchar_t *)javaTmpStg, wSource, wSourceNumChars);
        jStgUsed = wSourceNumChars;
    }

 	jDest = (*env)->NewString(env, javaTmpStg, jStgUsed);
  	if (javaTmpStg != NULL && javaTmpStg != tmpBuf) {
   		free(javaTmpStg);
   	}
   	javaTmpStg = NULL;
    retStr = (*env)->NewGlobalRef(env,jDest);
    (*env)->DeleteLocalRef(env, jDest);

    return(retStr);
}

char *getBasename(char *progStr)
{
	char *ptr = strrchr(progStr,(int)pathSepChar);
	if (ptr == (char *)NULL) {
		return (progStr);
	}
	return (++ptr);
}

char *truncateBasename(char *progStr)
{
	char *ptr = strrchr(progStr,(int)pathSepChar);
	if (ptr == (char *)NULL) {
		return (NULL);
	}
	*ptr = '\0'; // Truncate
	return (progStr);
}

void error(char *fmtStr, ...)
{
	va_list argPtr;
	va_start(argPtr, fmtStr);
	/* We can patch into an error mechanism here.  For now, go with fprintf */
	(void)vfprintf(stderr, fmtStr, argPtr);
	va_end(argPtr);
}

void debug(char *fmtStr, ...)
{
	va_list argPtr;
	va_start(argPtr, fmtStr);
	if (debugFlag) {
		/* We can patch into an error mechanism here.  For now, go with fprintf */
		(void)vfprintf(stdout, fmtStr, argPtr);
	}
	va_end(argPtr);
}

#if !(defined Convert32To16)
/*-------------------------------------------------------------------------------------------------
 * Surrogate Defines
 *-------------------------------------------------------------------------------------------------*/
#define SURROGATE_HIGH_START  0xD800   // Surrogate hi word start
#define SURROGATE_HIGH_END    0xDBFF   // "         "  "    end
#define SURROGATE_LOW_START   0xDC00   // "         low word start
#define SURROGATE_LOW_END     0xDFFF   // "         "   "    end

#define SURROGATE_SHIFT       10       // Shift 10 bits to convert from/to UTF32
#define SURROGATE_HALFBASE    0x10000  // UTF32 -> UTF16 split point for high order
#define SURROGATE_HALFMASK    0x3FF    // UTF32 -> UTF16 mask to get low order
/*-------------------------------------------------------------------------------------------------
 * UTF 32 <-> UTF 16 defines
 *-------------------------------------------------------------------------------------------------*/
#define MAX_UTF16             0xFFFF   // Max Basic Multilingual Plane (BMP) (Highest UTF16 char)
#define MAX_UTF16_32          0x10FFFF // Max 32 bit that can be converted to 16 bit (highest surrogate)



/*-------------------------------------------------------------------------------------------------
 * Convert32to16
 *    Converts UTF32 to UTF16  -- Stolen from ODPPPrvDataTypeS
 *-------------------------------------------------------------------------------------------------*/
int Convert32To16(const unsigned int *Utf32,
                     int Utf32Chars,
                     unsigned short *Utf16,
                     int Utf16Size,
                     int *pOutUtf16Size)
{
   int mErrorCode = RC_SUCCESS;
   unsigned int u32;
   int mInCharsUsed = 0;
   int mOutCharsUsed = 0;

   // Keep getting UTF32 data
   while (Utf32Chars)
   {
      u32 = *Utf32++;
      Utf32Chars--;
      mInCharsUsed++;

      // End of string
      if (!u32)
         break;

      // Will this fit as is?
      if (u32 <= MAX_UTF16)
      {
         // Still, if this is a high or low surrogate it is an error in UTF32
         if (u32 >= SURROGATE_HIGH_START && u32 <= SURROGATE_LOW_END)
         {
            mErrorCode = RC_FATAL; //EINVAL
            break;
         }

         // Ok, fits just find thanks, do it if not length request
         if (Utf16)
         {
            if (Utf16Size < 1)
            {
               mErrorCode = RC_FATAL; //ODPP_DATACNV_ERR_DATA_TRUNCATED;
               break;
            }

            *Utf16++ = (unsigned short)u32;
            Utf16Size--;
         }

         mOutCharsUsed++;
      }

      // Can this be converted to UTF16 with surrogates?
      else if (u32 > MAX_UTF16_32)
      {
         mErrorCode = RC_FATAL; //EINVAL;
         break;
      }

      // Put out surrogate
      else
      {
         // Check it out
         if (Utf16)
         {
            if (Utf16Size < 2)
            {
               mErrorCode = RC_FATAL; //ODPP_DATACNV_ERR_DATA_TRUNCATED;
               break;
            }

            Utf16Size -= 2;

            mOutCharsUsed++;

            // Between 0xFFFF -> 0x10FFF, convert to surrogate UTF16
            u32 -= SURROGATE_HALFBASE;

            // Put out high surrogate
            *Utf16++ = (unsigned short)( (u32 >> SURROGATE_SHIFT) + SURROGATE_HIGH_START );

            // Put out low surrogate
            *Utf16++ = (unsigned short)( (u32 & SURROGATE_HALFMASK) + SURROGATE_LOW_START );
         }

         // Length request
         else
         {
            mOutCharsUsed += 2;
         }
      }
   }

   if (Utf16)
      if (Utf16Size > 0)
         *Utf16 = 0;

   if(NULL != pOutUtf16Size)
      (*pOutUtf16Size) = mOutCharsUsed;

   return mErrorCode == 0 ? RC_SUCCESS : mErrorCode;
}
#endif
#ifndef ENV_VSTG_GET_REGION
#define ENV_VSTG_GET_REGION(pparea, len, pool, option) \
	(*(pparea) = malloc((len)), 0)
#define ENV_VSTG_FREE_REGION(parea, len, pool) ((void)free(parea),0)
#endif

/* This probably belongs in a helper file, rather than in a JNI file */
OptimReportingMessage_t *allocateMessagesArray(int numElements)
{
	OptimReportingMessage_t *allocatedBlock = NULL;
	int rc;
	/* Will return NULL on allocation failure */
	size_t allocSize = numElements * (sizeof(OptimReportingMessage_t) + sizeof(long long)); // Add extra for alignment
	rc = ENV_VSTG_GET_REGION(&allocatedBlock, allocSize, 0, 0);
	if (rc != 0) {
		allocatedBlock = NULL;
	}
	return(allocatedBlock);
}

void freeMessagesArray(OptimReportingMessage_t *allocatedBlock, int numElements)
{
	int rc = 0;
	size_t allocSize = numElements * (sizeof(OptimReportingMessage_t) + sizeof(long long)); // Add extra for alignment
	if (allocatedBlock == NULL) {
		return;
	}
	rc = ENV_VSTG_FREE_REGION(allocatedBlock, allocSize, 0);
	// Not sure what to do if rc != 0.  FAIL
}


void setMessageArrayElement(OptimReportingMessage_t *pBlock, int elementNumber, int id, wchar_t *message, int msgType)
{
	if (pBlock == (OptimReportingMessage_t *) NULL) {
		return; //Probably should scream bloody murder here.
	}
	if (message == (wchar_t *) NULL) {
		return; // Again, we should probably complain bitterly
	}
	pBlock[elementNumber].id = id;
	pBlock[elementNumber].message = message;
	pBlock[elementNumber].type = msgType;
	return;
}

/* Now do the same for RequestOptions */
OptimReportingNameValuePair_t *allocateRequestOptionsArray(int numElements)
{
	OptimReportingNameValuePair_t *allocatedBlock = NULL;
	int rc;
	/* Will return NULL on allocation failure */
	size_t allocSize = numElements * (sizeof(OptimReportingNameValuePair_t) + sizeof(long long)); // Add extra for alignment
	rc = ENV_VSTG_GET_REGION(&allocatedBlock, allocSize, 0, 0);
	if (rc != 0) {
		allocatedBlock = NULL;
	}
	return(allocatedBlock);
}

void freeRequestOptionsArray(OptimReportingNameValuePair_t *allocatedBlock, int numElements)
{
	int rc = 0;
	size_t allocSize = numElements * (sizeof(OptimReportingNameValuePair_t) + sizeof(long long)); // Add extra for alignment
	if (allocatedBlock == NULL) {
		return;
	}
	rc = ENV_VSTG_FREE_REGION(allocatedBlock, allocSize, 0);
	// Not sure what to do if rc != 0.  FAIL
}


void setRequestOptionsArrayElement(OptimReportingNameValuePair_t *pBlock, int elementNumber, wchar_t *name, wchar_t *value)
{
	if (pBlock == (OptimReportingNameValuePair_t *) NULL) {
		return; //Probably should scream bloody murder here.
	}
	if (name == (wchar_t *) NULL) {
		return; // Again, we should probably complain bitterly
	}
	if (value == (wchar_t *) NULL) {
		return; // Again, we should probably complain bitterly
	}
	pBlock[elementNumber].name = name;
	pBlock[elementNumber].value = value;
	return;
}

/* And for ExtendedObjects */
OptimReportingNumberPair_t *allocateExtendedObjectsArray(int numElements)
{
	OptimReportingNumberPair_t *allocatedBlock = NULL;
	int rc;
	/* Will return NULL on allocation failure */
	size_t allocSize = numElements * (sizeof(OptimReportingNumberPair_t) + sizeof(long long)); // Add extra for alignment
	rc = ENV_VSTG_GET_REGION(&allocatedBlock, allocSize, 0, 0);
	if (rc != 0) {
		allocatedBlock = NULL;
	}
	return(allocatedBlock);
}

void freeExtendedObjectsArray(OptimReportingNumberPair_t *allocatedBlock, int numElements)
{
	int rc = 0;
	size_t allocSize = numElements * (sizeof(OptimReportingNumberPair_t) + sizeof(long long)); // Add extra for alignment
	rc = ENV_VSTG_FREE_REGION(allocatedBlock, allocSize, 0);
	// Not sure what to do if rc != 0.  FAIL
}


void setExtendedObjectsArrayElement(OptimReportingNumberPair_t *pBlock, int elementNumber, int number1, int number2)
{
	if (pBlock == (OptimReportingNumberPair_t *) NULL) {
		return; //Probably should scream bloody murder here.
	}
	pBlock[elementNumber].number1 = number1;
	pBlock[elementNumber].number2 = number2;
	return;
}

/* And for Entities */
OptimReportingEntity_t *allocateEntitiesArray(int numElements)
{
	OptimReportingEntity_t *allocatedBlock = NULL;
	int rc;
	/* Will return NULL on allocation failure */
	size_t allocSize = numElements * (sizeof(OptimReportingEntity_t) + sizeof(long long)); // Add extra for alignment
	rc = ENV_VSTG_GET_REGION(&allocatedBlock, allocSize, 0, 0);
	if (rc != 0) {
		allocatedBlock = NULL;
	}
	return(allocatedBlock);
}

void freeEntitiesArray(OptimReportingEntity_t *allocatedBlock, int numElements)
{
	int rc = 0;
	size_t allocSize = numElements * (sizeof(OptimReportingEntity_t) + sizeof(long long)); // Add extra for alignment
	rc = ENV_VSTG_FREE_REGION(allocatedBlock, allocSize, 0);
	// Not sure what to do if rc != 0.  FAIL
}

/* Too many elements for a convenience setter on Entity... */
/* And for Relationships */
OptimReportingRelationship_t *allocateRelationshipsArray(int numElements)
{
	OptimReportingRelationship_t *allocatedBlock = NULL;
	int rc;
	/* Will return NULL on allocation failure */
	size_t allocSize = numElements * (sizeof(OptimReportingRelationship_t) + sizeof(long long)); // Add extra for alignment
	rc = ENV_VSTG_GET_REGION(&allocatedBlock, allocSize, 0, 0);
	if (rc != 0) {
		allocatedBlock = NULL;
	}
	return(allocatedBlock);
}

void freeRelationshipsArray(OptimReportingRelationship_t *allocatedBlock, int numElements)
{
	int rc = 0;
	size_t allocSize = numElements * (sizeof(OptimReportingRelationship_t) + sizeof(long long)); // Add extra for alignment
	rc = ENV_VSTG_FREE_REGION(allocatedBlock, allocSize, 0);
	// Not sure what to do if rc != 0.  FAIL
}


void setRelationshipsArrayElement(OptimReportingRelationship_t *pBlock, int elementNumber,
		wchar_t *relationshipName, wchar_t *parentName, wchar_t *childName, int parentAccessType, int parentKeyLimit,
		int childAccessType, int childKeyLimit)
{
	if (pBlock == (OptimReportingRelationship_t *) NULL) {
		return; //Probably should scream bloody murder here.
	}
	if (relationshipName == (wchar_t *) NULL) {
		return; // Again, we should probably complain bitterly
	}
	if (parentName == (wchar_t *) NULL) {
		return; // Again, we should probably complain bitterly
	}
	if (childName == (wchar_t *) NULL) {
		return; // Again, we should probably complain bitterly
	}
	pBlock[elementNumber].relationshipName = relationshipName;
	pBlock[elementNumber].parentName = parentName;
	pBlock[elementNumber].childName = childName;
	pBlock[elementNumber].parentAccessType = parentAccessType;
	pBlock[elementNumber].parentKeyLimit = parentKeyLimit;
	pBlock[elementNumber].childAccessType = childAccessType;
	pBlock[elementNumber].childKeyLimit = childKeyLimit;
	return;
}

/* And for Statistics */
OptimReportingStatistics_t *allocateStatisticsArray(int numElements)
{
	OptimReportingStatistics_t *allocatedBlock = NULL;
	int rc;
	/* Will return NULL on allocation failure */
	size_t allocSize = numElements * (sizeof(OptimReportingStatistics_t) + sizeof(long long)); // Add extra for alignment
	rc = ENV_VSTG_GET_REGION(&allocatedBlock, allocSize, 0, 0);
	if (rc != 0) {
		allocatedBlock = NULL;
	}
	return(allocatedBlock);
}

void freeStatisticsArray(OptimReportingStatistics_t *allocatedBlock, int numElements)
{
	int rc = 0;
	size_t allocSize = numElements * (sizeof(OptimReportingStatistics_t) + sizeof(long long)); // Add extra for alignment
	rc = ENV_VSTG_FREE_REGION(allocatedBlock, allocSize, 0);
	// Not sure what to do if rc != 0.  FAIL
}
/* Too many elements for a convenience setter on Statistics... */
/* Frees all the internall allocated elements of a context, since they should be allocated with our convenience methods. */
void freeContextInternals(OptimReportingContext_t *context)
{
	freeEntitiesArray(context->entities, context->numEntities);
	context->entities = (OptimReportingEntity_t *)NULL;
	context->numEntities = 0;

	freeExtendedObjectsArray(context->extendedObjects, context->numExtendedObjects);
	context->extendedObjects = (OptimReportingNumberPair_t *)NULL;
	context->numExtendedObjects = 0;

	freeMessagesArray(context->messages, context->numMessages);
	context->messages = (OptimReportingMessage_t *)NULL;
	context->numMessages = 0;

	freeRelationshipsArray(context->relationships, context->numRelationships);
	context->relationships = (OptimReportingRelationship_t *)NULL;
	context->numRelationships = 0;

	freeRequestOptionsArray(context->requestOptions, context->numRequestOptions);
	context->requestOptions = (OptimReportingNameValuePair_t *)NULL;
	context->numRequestOptions = 0;

	freeStatisticsArray(context->statistics, context->numStatistics);
	context->statistics = (OptimReportingStatistics_t *)NULL;
	context->numStatistics = 0;
}
/* Done with Array convenience routines */


void populateTestContext(OptimReportingContext_t *context)
{
	context->reportFile=L"/tmp/ReportTest.xml";
	context->htmlReportFile=L"/tmp/ReportTest.html";
	context->reportType = ORC_ARCHIVE_REPORT;
	context->optimDirectoryName = L"DIR_ORA112";
	context->optimDirectoryDBMSType = L"Oracle";
	context->optimDirectoryDBMSVersion = L"11.2";
	context->optimDirectoryDBMSVendor = L"Oracle";
	context->optimSolutionName = L"Optim Data Growth";
	context->optimSolutionVersion = L"9.1.0.5";
	context->optimRuntimeName = L"Optim Distributed";
	context->optimRuntimeVersion = L"9.1.0.5";
	context->clientUserId = L"optimuser";
	context->errorCount = 1;
	context->infoCount = 3;
	context->requestName = L"OPTIMUSER_ARCHIVE1";
	context->serverName = L"(local)";
	context->serverUserId = L"optimuser";
	context->timeElapsed = 12345L;
	context->timeStarted = L"06-17-2013 13:24:00";
	context->timeFinished = L"06-17-2013 13:27:00";
	context->warningCount = 0;
	/* First instance of an array type of thing... */
	context->numMessages = 2;
	/* Not checking for NULL from allocation... this is test/demo code, not production */
	context->messages = allocateMessagesArray(context->numMessages);
	setMessageArrayElement(context->messages, 0, 0, L"For those about to rock", ORM_MESSAGE_TYPE_ERROR);
	setMessageArrayElement(context->messages, 1, 1, L"We salute you", ORM_MESSAGE_TYPE_WARNING);
	context->totalRows = 90;
	context->totalRowsFirstPass = 30;
	context->totalRowsInError = 1;
	context->totalTables = 2;
	context->numRequestOptions = 1;
	context->requestOptions = allocateRequestOptionsArray(context->numRequestOptions);
	setRequestOptionsArrayElement(context->requestOptions, 0, L"RequestOption 1", L"Splunge");
	context->extractFile = L"/opt/IBM/InfoSphere/Optim/ArchiveFile.af";
	context->accessDefinitionName = L"OPTIMUSER_AD";
	context->accessDefinitionType = ORC_ACCESS_DEFINITION_TYPE;
	context->accessDefinitionModificationDate = L"2013-03-24";
	context->numExtendedObjects = 1;
	context->extendedObjects = allocateExtendedObjectsArray(context->numExtendedObjects);
	setExtendedObjectsArrayElement(context->extendedObjects, 0, 0, ORC_ALIAS_TYPE);
	context->numEntities = 2;
	context->entities = allocateEntitiesArray(context->numEntities);
	/* Set everything for 1st entity */
	context->entities[0].failed = 0;
	context->entities[0].name = L"OPTIM.OPTIM_PARENT";
	context->entities[0].recordCount = 30L;
	context->entities[0].skipped = 0;
	context->entities[0].status = L" ";
	context->entities[0].type = ORE_TABLE_TYPE;
	context->entities[0].extractAverageRowLength = 255.4;
	context->entities[0].extractLob = TRUE;
	context->entities[0].extractMaxLobLength = 123456789.0;
	context->entities[0].extractCompressionEnabled = FALSE;
	context->entities[0].extractCompressionValue = 0;
			/* Next Entity */
	context->entities[1].failed = 3;
	context->entities[1].name = L"OPTIM.OPTIM_CHILD";
	context->entities[1].recordCount = 40L;
	context->entities[1].skipped = 0;
	context->entities[1].status = L"A-OK";
	context->entities[1].type = ORE_TABLE_TYPE;
	context->entities[1].extractAverageRowLength = 255.4;
	context->entities[1].extractLob = TRUE;
	context->entities[1].extractMaxLobLength = 123456789.0;
	context->entities[1].extractCompressionEnabled = TRUE;
	context->entities[1].extractCompressionValue = 3;
	/* Relationships */
	context->numRelationships = 1;
	context->relationships = allocateRelationshipsArray(context->numRelationships);
	setRelationshipsArrayElement(context->relationships, 0, L"OPTIM.MOTHER_AND_CHILD_REUNION",
			L"OPTIM.OPTIM_PARENT", L"OPTIM.OPTIM_CHILD", ORR_SCAN_STRATEGY, 0, ORR_SCAN_STRATEGY, 0);
	/* Statistics */
	context->numStatistics = 1;
	context->statistics = allocateStatisticsArray(context->numStatistics);
	context->statistics[0].childStrategy = ORS_NO_KEYS_STRATEGY;
	context->statistics[0].columnCount = 0;
	context->statistics[0].cycle = TRUE;
	context->statistics[0].dbAccessType = ORS_SCAN;
	context->statistics[0].dbDataPerSecond = 0;
	context->statistics[0].dbKeysPer = 0;
	context->statistics[0].dbOpenCursor = 0;
	context->statistics[0].dbProcessTime = 12345L;
	context->statistics[0].dbRowsFetched = 1457;
	context->statistics[0].dbRowsPerSecond = 1450;
	context->statistics[0].dbRowsWritten = 1457;
	context->statistics[0].dbTimeInAct = 12300L;
	context->statistics[0].dbTimeInDB = 45L;
	context->statistics[0].estimatedRowCount = 0;
	context->statistics[0].fetchBuffer = 0;
	context->statistics[0].foreignKeyIndex = 0;
	context->statistics[0].foreignKeyNoIndex = 0;
	context->statistics[0].lobs = TRUE;
	context->statistics[0].method = ORS_PRIMARY_KEY_METHOD;
	context->statistics[0].name = L"Statistic Name";
	context->statistics[0].parentStrategy = ORS_NO_KEYS_STRATEGY;
	context->statistics[0].primaryKeyIndex = 0;
	context->statistics[0].primaryKeyNoIndex = 0;
	context->statistics[0].relAccess = ORS_FORCING_ACCESS;
	context->statistics[0].relDb2Cost = 0;
	context->statistics[0].relDb2ScanCost = 0;
	context->statistics[0].relDirection = ORS_PARENT_DIRECTION;
	context->statistics[0].relIndex = TRUE;
	context->statistics[0].relKeyLength = 0;
	context->statistics[0].relKeyPerCur = 0;
	context->statistics[0].relLookupKey = 0;
	context->statistics[0].relName = L"OPTIM.MOTHER_AND_CHILD_REUNION";
	context->statistics[0].relSql = L"SELECT count(FOO) FROM OPTIM.OPTIM_CUSTOMERS";
	context->statistics[0].relType = ORS_REL_TYPE_DATABASE;
	context->statistics[0].rowLength = 0;
	context->statistics[0].selectUniqueRecord = TRUE;
	context->statistics[0].stepNumber = 0;
	context->statistics[0].dbmsVendor = L"Oracle";
	context->statistics[0].dbmsVersion = L"11.2";
}

int
main(int argc, char **argv)
{
	OptimReportingContext_t context;
	int rc = RC_SUCCESS;

	(void)printf("Starting report test\n");
	/* Populate C Context */
	/* Use /tmp/ReportTest.xml and /tmp/ReportTest.html */
	populateTestContext(&context);
	(void)printf("Populated Test context\n");

	/* Call createReport */
	(void)printf("Outputting Report\n");
	rc = outputReport(&context);
	(void)printf("Return Code %d, now freeing internals\n", rc);
	/* Deallocate all */
	freeContextInternals(&context);
	(void)printf("Terminating...\n");
	/* End */
	return rc;
}
